#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FAL command for a cloud service."""

import libfal.falcmd as falcmd
import fal_config as config
import pal_info
from iobase.xmldata import (ANY_CHILD_PATH,)
from libfal.util import (servicectrl_suspend, servicectrl_resume)
from ionas.request import (Entry,)
from fal_info import (ReqTypes, Results,)
from fal_config import (FUNC_TO_SERVICE_MAP,)


SYNCCONFLIST_NAME = "syncconflist"
SYNCCONF_NAME = "syncconf"
ID_NAME = "id"
ID_MIN = 0
IGNORELIST_PATH = "ignorelist"
GET_ID_XPATH = SYNCCONFLIST_NAME + "/" + SYNCCONF_NAME + "/id"
FIND_BY_ID_XPATH_TMPL = SYNCCONFLIST_NAME + "/" + SYNCCONF_NAME + \
                            '/id[text()="%s"]/..'

# For cloud packages
NASDSYNC2_SERVICE = "nasdsync2"
# Add the service entry for the clone package
FUNC_TO_SERVICE_MAP[NASDSYNC2_SERVICE] = NASDSYNC2_SERVICE


class FalCloud(falcmd.FalCmd):
    """FAL Cloud command implementation class."""

    def __load_config(self, file_name):
        """Load a 'file_name' file setting and return XML data ."""

        return self.send_entry_with_file(Entry(ReqTypes.LOAD),
                                        pal_info.PAL_CONFFILE_PATH,
                                        file_name).get()

    def __save_config(self, file_name, xml_data):
        """Save the "xml_data" XML data to a 'file_name' file."""

        entry = Entry(ReqTypes.SAVE)
        entry.set(xml_data)
        return self.send_entry_with_file(entry,
                                pal_info.PAL_CONFFILE_PATH, file_name)

    def __add_handler(self, entry):
        """Handle add request and set entry to result."""

        added_node = entry.get().get_node(SYNCCONF_NAME)
        try:
            # Force delete ID tag.
            added_node.delete_children(ID_NAME)
        except:
            # Ignore error.
            pass

        sync_conf = self.__load_config(config.SYNC_CONF_FILE)

        id_list = [int(id_str) for id_str in
                        sync_conf.get_value_list(GET_ID_XPATH)]
        target_id = ID_MIN
        while True:
            if target_id in id_list:
                target_id += 1
                continue
            break

        syncconflist = sync_conf.get_node(SYNCCONFLIST_NAME)
        added_node.append_child(ID_NAME, str(target_id))
        syncconflist.append_child(added_node)

        servicectrl_suspend([], [NASDSYNC2_SERVICE])
        self.__save_config(config.SYNC_CONF_FILE, sync_conf)
        servicectrl_resume([], [NASDSYNC2_SERVICE])
        entry.set_result(Results.OK)
        return entry

    def __modify_handler(self, entry):
        """Handle modify request and set entry to result."""

        modified_node = entry.get().get_node(SYNCCONF_NAME)
        sync_conf = self.__load_config(config.SYNC_CONF_FILE)
        target_node = sync_conf.get_node(FIND_BY_ID_XPATH_TMPL %
                            (modified_node.get_node(ID_NAME).get_text(),))
        target_node.delete_children(ANY_CHILD_PATH)
        target_node.update(modified_node)

        servicectrl_suspend([], [NASDSYNC2_SERVICE])
        self.__save_config(config.SYNC_CONF_FILE, sync_conf)
        servicectrl_resume([], [NASDSYNC2_SERVICE])
        entry.set_result(Results.OK)
        return entry

    def __delete_handler(self, entry):
        """Handle delete request and set entry to result."""

        deleted_node = entry.get().get_node(SYNCCONF_NAME)
        sync_conf = self.__load_config(config.SYNC_CONF_FILE)

        syncconflist = sync_conf.get_node(SYNCCONFLIST_NAME)

        idx = 0
        for syncconf in syncconflist.get_node_list(SYNCCONF_NAME):
            if (syncconf.get_value(ID_NAME) !=
                    deleted_node.get_value(ID_NAME)):

                idx += 1
                continue
            syncconflist.delete_child(SYNCCONF_NAME, index=idx)

        servicectrl_suspend([], [NASDSYNC2_SERVICE])
        self.__save_config(config.SYNC_CONF_FILE, sync_conf)
        servicectrl_resume([], [NASDSYNC2_SERVICE])
        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        falcmd.FalCmd.__init__(self, option=option)

        save_and_load_handler = lambda entry: self.send_entry_with_file(
            entry, pal_info.PAL_CONFFILE_PATH, config.SYNC_CONF_FILE)

        self.set_handler(ReqTypes.SAVE, save_and_load_handler)
        self.set_handler(ReqTypes.LOAD, save_and_load_handler)
        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)

#
# Main
#

if __name__ == "__main__":
    FalCloud(option=falcmd.FalCmdOption()).handle_request()
