# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""GoogleDrive File Access module.

Define classes for OneDrive file access.
"""

# common module
from __future__ import unicode_literals
from fileaccess import (FileObj, Access)
import functools
import os
import stat
import arrow
import re
import time
from buffer_object import BufferObject, BufferObjectW, ChunkedBuffer
from fileaccess_helper import *
from pathid import (PathIdMap,)

# feature module
import httplib2
import oauth2client
from oauth2client.file import Storage
from oauth2client.client import OAuth2WebServerFlow
from googleapiclient.discovery import build
# from googleapiclient.http import MediaIoBaseDownload, DEFAULT_CHUNK_SIZE as GOOGLE_DEFAULT_CHUNK_SIZE
from googleapiclient.http import MediaIoBaseDownload
from googleapiclient.http import MediaIoBaseUpload
from googleapiclient.errors import HttpError
import mimetypes
import proxy_info
import googleapiclient_helper

# common constant
FILE_SEPARATOR = "/"
FILE_MODE = 0
DIR_MODE = stat.S_IFDIR
ERR_MSG_RENAME = "Invalid new name"
ERR_MSG_INVALID_FILE_LENGTH = "Invalid file length"
ERR_MSG_INVALID_OPEN_MODE = "Invalid open mode"
OPEN_MODE_READ = "r"
OPEN_MODE_WRITE = "w"

# feature constant
DIR_MIME_TYPE = "application/vnd.google-apps.folder"

SESSION_PATH_PERM = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP |
                        stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)

MAX_RETRY_COUNT = 5
RETRY_WAIT_TIME = 5


def _retry_ratelimitexceeded(max_retry_count=MAX_RETRY_COUNT):
    def _deco(func):
        @functools.wraps(func)
        def _func(*args, **kwargs):
            for count in range(max_retry_count):
                try:
                    return func(*args, **kwargs)
                except HttpError as e:
                    if e._get_reason() in ['Rate Limit Exceeded']:  # TODO
                        time.sleep(2**count)
                        continue
                    raise
        return _func
    return _deco


class GoogleDriveFileObj(FileObj):
    """Class for GoogleDrive file."""

    def __init__(self, access, path, mode, length, chunk_size):
        """Constructor."""
        super(GoogleDriveFileObj, self).__init__()
        self._access = access
        self._path = path
        self._mode = (filter((lambda m: m in mode.lower()), (OPEN_MODE_READ, OPEN_MODE_WRITE)) or [None])[0]
        self._length = length
        self._chunk_size = chunk_size
        if self._chunk_size is None:
            # self._chunk_size = GOOGLE_DEFAULT_CHUNK_SIZE
            self._chunk_size = 52428800
        self._r_pos = 0
        self._w_pos = 0
        self._downloader = None
        self._uploader = None

        if self._mode == OPEN_MODE_READ:
            self._download()
        elif self._mode == OPEN_MODE_WRITE:
            if self._length < 0:
                raise ValueError(ERR_MSG_INVALID_FILE_LENGTH)
            self._upload_prepare()
        else:
            raise ValueError(ERR_MSG_INVALID_OPEN_MODE)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        """Close file."""
        try:
            if self._mode == OPEN_MODE_WRITE:
                if self._buffer is not None and self._buffer.rest > 0:
                    self._upload(force=True)
        finally:
            self._buffer.close()

    def read(self, size=-1):
        """Return a "size" bytes data from current position in file.
        If the size argument is negative or omitted, then return all data.
        """

        if size < 0:
            size = self._length

        r_size = min(size, self._length - self._r_pos)
        ret_buf = BufferObjectW(size=r_size)
        while ret_buf.lack > 0:
            ret_buf.write(self._buffer.read(ret_buf.lack))
            if self._buffer.rest == 0:
                self._download()

        self._r_pos += ret_buf.length
        ret = ret_buf.read()
        ret_buf.close()
        return ret

    def _download(self):
        """Download to buffer"""

        if not self._downloader:
            item_id = self._access._get_item_id(self._path, be_raise=True)
            self._length = self._access.stat_file(self._path)[stat.ST_SIZE]
            request = self._access._service.files().get_media(
                fileId=item_id,
            )
            self._buffer = BufferObject()
            self._downloader = MediaIoBaseDownload(self._buffer, request, chunksize=min(self._chunk_size, self._length))

        self._buffer.clear()
        # status, done = self._downloader.next_chunk()
        _next_chunk = _retry_ratelimitexceeded()(self._downloader.next_chunk)
        status, done = _next_chunk()

        if done:
            def dummy():
                self._buffer.close()
                return None, True
            self._downloader.next_chunk = dummy

    def write(self, data):
        """Write a "data" data to current position in file."""

        self._buffer.write(data)
        self._w_pos += len(data)
        self._upload(force=False)

    def _upload_prepare(self):
        self._buffer = ChunkedBuffer(
            size=self._length,
            buffer_size=self._chunk_size,
            read_at_first=True,
            retry_limit=0,
        )
        mimetype = mimetypes.guess_type(self._path)
        if mimetype[0] is None:
            mimetype = ("application/octet-stream", None)
        self._buffer.set_r_mode()
        _resumable = True if self._length > 0 else False
        media = MediaIoBaseUpload(
            self._buffer,
            mimetype=mimetype[0],
            chunksize=self._chunk_size,
            resumable=_resumable)

        item_id = self._access._get_item_id(self._path)
        if item_id is not None:
            # update
            self._uploader = self._access._service.files().update(
                fileId=item_id,
                media_body=media,
            )
        else:
            # create
            name = os.path.basename(self._path)
            parent = os.path.dirname(self._path)
            parent_id = self._access._get_item_id(parent, be_raise=True)
            metadata = {
                'name': name,
                'parents': [parent_id],
                'mimeType': mimetype,
            }
            self._uploader = self._access._service.files().create(
                media_body=media,
                body=metadata,
            )
        self._buffer.set_r_mode()
        self._buffer.seek(0)

    def _upload(self, force=False, max_retry_count=10):
        if self._buffer.lack == 0 or force:
            self.retries = 0
            while True:
                try:
                    self._buffer.set_r_mode()
                    if self._length > 0:
                        status, response = self._uploader.next_chunk()
                    else:
                        response = self._uploader.execute()
                    self._buffer.close_chunk()
                    break
                except Exception as e:
                    import sys
                    info = sys.exc_info()
                    if self.retries > max_retry_count:
                        raise
                    else:
                        self.retries += 1
                        time.sleep(self.retries**2)


class GoogleDriveAccess(Access):
    """class accessing to GoogleDrive file."""

    def __init__(self,
                 client_id=None,
                 client_secret=None,
                 scopes=None,
                 redirect_url='http://localhost:8080',
                 session_path=None,
                 auth_url=oauth2client.GOOGLE_AUTH_URI,
                 token_url=oauth2client.GOOGLE_TOKEN_URI,
                 revoke_url=oauth2client.GOOGLE_REVOKE_URI,
                 proxy=None,
                 ):
        """Constructor.

        Args:
            client_id (unicode): client identifier.
            client_secret (unicode): client secret.
            scopes (unicode or, list of unicode): scope(s) of the credentials
                being requested.
            redirect_url (unicode): Either the string 'urn:ietf:wg:oauth:2.0:oob'
                for a non-web-based application, or a URI that
                handles the callback from the authorization server.
            session_path (unicode):
                File path for session(token, expired, etc...)
            auth_url (unicode):  URI for authorization endpoint. For convenience
                defaults to Google's endpoints but any OAuth 2.0 provider
                can be used.
            token_url (unicode): URI for token endpoint. For convenience
                defaults to Google's endpoints but any OAuth 2.0
                provider can be used.
            revoke_url (unicode): URI for revoke endpoint. For convenience
                defaults to Google's endpoints but any OAuth 2.0
                provider can be used.
            proxy (unicode):
                Mapping of protocols to proxy URLs.
                If None, bypass proxies
                e.g.
                proxy = 'http://127.0.0.1:8888'
        """

        super(GoogleDriveAccess, self).__init__()

        self._flow = OAuth2WebServerFlow(
            client_id=client_id,
            client_secret=client_secret,
            scope=scopes,
            redirect_uri=redirect_url,
            auth_uri=auth_url,
            token_uri=token_url,
            revoke_uri=revoke_url,
            access_type='offline',
            prompt='consent',
        )

        self._service = None
        self._storage = Storage(session_path)
        self._session_path = session_path

        self._credentials = None
        if os.path.exists(session_path):
            self._credentials = self._storage.get()

        if not self._credentials or self._credentials.invalid:
            self._auth_url = self._flow.step1_get_authorize_url()
        else:
            self._auth_url = redirect_url

        self._proxy = proxy
        self._path_id_map = PathIdMap()
        self._path_id_map.append_path(FILE_SEPARATOR, "root")

    @property
    def auth_url(self):
        """Get the auth url using the params provided
        If session is already valid, returns redirect_url
        """
        return self._auth_url

    def authenticate(self, params=None):
        """Takes in params, creates a session.

        Args:
            params(dict):
                The GET parameter sent to redirect_url
                If session is already valid, is None.
                e.g.
                {'code': '....'}
        """

        if params is None:
            params = {}

        # if self._proxy is None:
        #     http_provider = httplib2.Http()
        # else:
        #     pi = proxy_info.proxy_info_from_url(self._proxy)
        #     http_provider = httplib2.Http(proxy_info=pi)
        def _get_http(pi=None if self._proxy is None else proxy_info.proxy_info_from_url(self._proxy)):
            return httplib2.Http(proxy_info=pi) if pi else httplib2.Http()
        http_provider = _get_http()

        if params.get('code') is not None:
            self._credentials = self._flow.step2_exchange(params['code'], http_provider)
            self._storage.put(self._credentials)
            if os.path.exists(self._session_path):
                try:
                    os.chmod(self._session_path, SESSION_PATH_PERM)
                except:
                    pass

        # self._service = build('drive', 'v3',
        #                       http=self._credentials.authorize(http_provider))
        import googleapiclient
        def _builder(http, *args, **kwargs):
            new_http = _get_http()
            new_http = self._credentials.authorize(new_http)
            return googleapiclient.http.HttpRequest(new_http, *args, **kwargs)
        self._service = build(
            'drive',
            'v3',
            # http=self._credentials.authorize(http_provider),
            requestBuilder=_builder,
        )

        # self._storage.put(self._credentials)
        # if os.path.exists(self._session_path):
        #     try:
        #         os.chmod(self._session_path, SESSION_PATH_PERM)
        #     except:
        #         pass

    def _stat(self, target_path=None, target_item=None):
        """Return a stat information dictionary for "target_path"."""

        if target_item is None:
            item_id = self._get_item_id(target_path, be_raise=True)
            retry_count = 0
            while True:
                try:
                    target_item = self._service.files().get(
                        fileId=item_id,
                        fields="mimeType, modifiedTime, size"
                    ).execute()
                    break
                except HttpError, exc:
                    if (exc._get_reason() != "Rate Limit Exceeded" or
                        retry_count >= MAX_RETRY_COUNT):

                        raise
                    retry_count += 1
                    time.sleep(RETRY_WAIT_TIME)

        _mode = FILE_MODE
        if target_item.get("mimeType") == DIR_MIME_TYPE:
            _mode = DIR_MODE

        _mtime = target_item.get("modifiedTime")
        if _mtime is not None:
            _mtime = arrow.get(_mtime).timestamp

        _ctime = None
        _size = None
        if _mode == FILE_MODE:
            _size = target_item.get("size")
            if _size is not None:
                _size = int(_size)

        return {
            stat.ST_MODE: _mode,
            stat.ST_MTIME: _mtime,
            stat.ST_CTIME: _ctime,
            stat.ST_SIZE: _size,
        }

    def _get_directory(self, path, be_raise=False):
        return self._get_item_id(path, be_raise)

    def _get_file(self, path, be_raise=False):
        return self._get_item_id(path, be_raise)

    def _get_item_id(self, path, be_raise=False):
        if path == FILE_SEPARATOR or path == "":
            return "root"

        try:
            target_id = self._path_id_map.get_id(path)
            if target_id is not None:
                return target_id
        except:
            pass

        dir_name = os.path.dirname(path)
        file_name = os.path.basename(path)

        parent_id = self._get_item_id(dir_name)
        if parent_id is None:
            return None

        file_name = re.sub("([^\\\\]|^)'", "\\1\\'", file_name)

        q = "trashed=false and name='%s' and '%s' in parents" % (file_name, parent_id)

        retry_count = 0
        while True:
            try:
                response = self._service.files().list(
                    q=q,
                    spaces="drive",
                    fields='nextPageToken, files(id, name, version, modifiedTime, mimeType, size)',
                ).execute()
                break
            except HttpError, exc:
                if (exc._get_reason() != "Rate Limit Exceeded" or
                    retry_count >= MAX_RETRY_COUNT):

                    raise
                retry_count += 1
                time.sleep(RETRY_WAIT_TIME)

        items = response.get('files', [])
        items = sorted(items, key=lambda x:(-int(x["version"]), -arrow.get(x["modifiedTime"]).timestamp))
        if len(items) > 0:
            item = items[0]
            self._path_id_map.append_path(path, item["id"])
            return item["id"]
        elif len(items) == 0:
            if be_raise:
                raise LookupError(path)
            return None

    def _get_item_list(self, parent_name="", parent_id=None, recursive=False):
        ret = []
        q = "trashed=false"
        page_token = None
        if parent_id:
            q += " and '%s' in parents" % parent_id

        while True:
            retry_count = 0
            while True:
                try:
                    response = self._service.files().list(
                        q=q,
                        spaces="drive",
                        fields='nextPageToken, files(id, name, version, modifiedTime, mimeType, size)',
                        pageToken=page_token,
                    ).execute()
                    break
                except HttpError, exc:
                    if (exc._get_reason() != "Rate Limit Exceeded" or
                        retry_count >= MAX_RETRY_COUNT):

                        raise
                    retry_count += 1
                    time.sleep(RETRY_WAIT_TIME)

            for item in response.get('files', []):
                if parent_name == FILE_SEPARATOR:
                    item_name = "%s/%s" % ("", item['name'])
                else:
                    item_name = "%s/%s" % (parent_name, item['name'])
                item_id = item["id"]
                item["name"] = item_name
                ret.append(item)
                if recursive and item['mimeType'] == "application/vnd.google-apps.folder":
                    ret.extend(
                        self._get_item_list(
                            parent_name=item_name,
                            parent_id=item_id))
            page_token = response.get('nextPageToken', None)
            if page_token is None:
                break
        return ret

    def _delete_item(self, target_path=None, target_id=None):
        if target_id is None:
            target_id = self._get_item_id(target_path, be_raise=True)

        retry_count = 0
        while True:
            try:
                self._service.files().delete(fileId=target_id).execute()
                self._path_id_map.delete_path(target_path)
                break
            except HttpError, exc:
                if (exc._get_reason() != "Rate Limit Exceeded" or
                    retry_count >= MAX_RETRY_COUNT):

                    raise
                retry_count += 1
                time.sleep(RETRY_WAIT_TIME)

    def _rename_item(self, target_path=None, target_id=None, new_name=None):
        if target_id is None:
            target_id = self._get_item_id(target_path, be_raise=True)

        data = {
            'name': new_name,
        }

        self._service.files().update(
            fileId=target_id,
            body=data,
        ).execute()

    def _move_item(self, target_path=None, target_id=None, new_dir_path=None):
        if target_id is None:
            target_id = self._get_item_id(target_path, be_raise=True)

        old_parent_path = os.path.dirname(target_path)
        old_parent_id = self._get_item_id(old_parent_path)
        new_dir_id = self._get_item_id(new_dir_path, be_raise=True)

        self._service.files().update(
            fileId=target_id,
            addParents=new_dir_id,
        ).execute()
        self._service.files().update(
            fileId=target_id,
            removeParents=old_parent_id,
        ).execute()

    def make_directory(self, dir_path):
        """Create "dir_path" directory."""

        if dir_path == "" or dir_path == FILE_SEPARATOR:
            # if blank, do nothing
            return

        target_id = self._get_item_id(dir_path)
        if target_id is not None:
            # if exists, do nothing
            return

        parent_path = os.path.dirname(dir_path)
        if parent_path == FILE_SEPARATOR or parent_path == "":
            parent_id = None
        else:
            parent_id = self._get_item_id(parent_path)
            if parent_id is None:
                self.make_directory(parent_path)
                parent_id = self._get_item_id(parent_path)

        target_name = os.path.basename(dir_path)

        metadata = {
            'name': target_name,
            'mimeType': DIR_MIME_TYPE,
        }
        if parent_id is not None:
            metadata['parents'] = [parent_id]

        retry_count = 0
        while True:
            try:
                self._service.files().create(
                    body=metadata,
                ).execute()
                break
            except HttpError, exc:
                if (exc._get_reason() != "Rate Limit Exceeded" or
                    retry_count >= MAX_RETRY_COUNT):

                    raise

                retry_count += 1
                time.sleep(RETRY_WAIT_TIME)

    def delete_directory(self, dir_path):
        """Delete "dir_path" directory."""

        self._delete_item(target_path=dir_path)

    def stat_directory(self, dir_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        return self._stat(target_path=dir_path)

    def rename_directory(self, dir_path, new_name):
        """Rename a name of "dir_path" to "new_name"."""

        if FILE_SEPARATOR in new_name:
            raise Exception("Invalid new_name", {"new_name": new_name})

        self._rename_item(target_path=dir_path, new_name=new_name)

    def move_directory(self, dir_path, new_dir_path):
        """Move "dir_path" to "new_path"."""

        self._move_item(target_path=dir_path,
                        new_dir_path=new_dir_path)

    def get_contents(self, dir_path, with_stat=False):
        """Return a contents dictionary in "dir_path".
           A stat information must be included, if "with_stat" is True,
        """

        items = self._get_item_list(
            parent_name=dir_path,
            parent_id=self._get_item_id(dir_path, be_raise=True))
        items = sorted(items, key=lambda x: (-int(x["version"]), -arrow.get(x["modifiedTime"]).timestamp))

        contents = {}
        for item in items:
            _name = os_path_bottom_name(item["name"])
            if _name not in contents:
                if with_stat:
                    contents[_name] = self._stat(target_item=item)
                else:
                    contents[_name] = {}

        return contents

    def get_updated_contents(self, dir_path):
        """Return a updated contents dictionary in "dir_path"."""

        raise NotImplementedError

    def open(self, file_path, mode="r", length=-1, chunk_size=None):
        """Return a class instance of the "FileObj" or sub class
           corresponding to "file_path".
        """

        return GoogleDriveFileObj(self, file_path, mode, length, chunk_size)

    def delete_file(self, file_path):
        """Delete "file_path" file."""

        self._delete_item(target_path=file_path)

    def stat_file(self, file_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        return self._stat(target_path=file_path)

    def rename_file(self, file_path, new_name):
        """Rename a name of "file_path" to "new_name"."""

        if FILE_SEPARATOR in new_name:
            raise Exception("Invalid new_name", {"new_name": new_name})

        self._rename_item(target_path=file_path, new_name=new_name)

    def move_file(self, file_path, new_dir_path):
        """Move "file_path" to "new_path"."""

        self._move_item(target_path=file_path,
                        new_dir_path=new_dir_path)

    def test(self):
        pass
